/* "cal2unix.c" 
 *
 * A program to extract information from a MS Windows Calendar file.
 *
 * Copyright 1994 Thomas Insel (tinsel@uiuc.edu)
 * 210 Parkview Drive, Bloomington IL 61701-2038
 */

/**** NOTES **

There is probably no easy way to make this a pipe, since we make heavy
usage of fseek().  We could use more byte counting, but the specs
don't guaratee that entries are in order.

Usage message & command line options
 -- force 24/12 hour format

Also, design a good output format.  My current favorite:

15 Apr 1994 (x_ Mail Income Tax forms
15 Apr 1995 12:00 Lunch with Fred
...

*************/

#include <stdio.h>
#include <time.h>

#define SECS_PER_DAY 60*60*24 /* number of seconds in one day */

/* I don't know anything about MS Windows programming, but these seem
 * to be the logical typedefs.  These are architecture-specific, and
 * the entire program will require adjustment if the byte-order
 * changes */

/* 8-bit data types */
typedef unsigned char		winByte;

/* 16-bit data types */
typedef unsigned short 		winWord;
typedef unsigned short      winUint;
typedef short				winInt;
typedef short				winBool;

/* These data structures for the .CAL file are adapted from the book
 * _Inside_Windows_File_Formats_ by Tom Swan (SAMS Publishing, 1993). */

/* The first 64 bytes of the file are this header. */
typedef struct xcalHeader {
    char 		signature[8];
    winWord		dateDescriptors;  /* Number of 64-byte dateInfo records, 
                                  /* which follow immediately. */
    /* These are preference settings, which we mostly ignore. */
    winWord		minEarlyRing;
	winBool		fSound; 
	winInt		interval;
	winInt		minInterval;
	winBool		f24HourFormat;
	winInt		startTime;

	winWord		reserved[21];
} calHeader;

typedef struct xdateInfo {
    winUint		date;
    winWord		markSymbol;
    winWord		numAlarms;   /* This is not the number of apptInfo
                              * records, so we ignore it. */
    winUint		fileBlockOffset;  /* Offset to this dates dayInfo record 
                                   * in 64-byte units.  There is no dayInfo 
                                   * record if this is set to 0xffff. */
    winWord		reserved1;
    winInt		reserved2;
} dateInfo;

typedef struct xdayInfo {
    winUint		reserved1;
    winUint		date;
    winUint		reserved2;
    winUint		cbNotes; 
    winUint		cbAppointments;
    /* cbNotes bytes of notes go here, followed by cbAppointments
     * bytes of apptInfo records. */
} dayInfo;

typedef struct xapptInfo {
    winByte		apptSize;  /* size of the entire record */
    winByte		flags;
    winWord		time;
    /* (apptSize - 4) bytes of description go here. */
} apptInfo;

/* Verify that this is actually a MS Windows Calendar file.  These magic 
 * numbers are from BAND("CALENDAR", "radnelac"). */
char checkSignature(unsigned char *sig) {
     return (sig[0] == 0xb5) &&
            (sig[1] == 0xa2) &&
            (sig[2] == 0xb0) &&
            (sig[3] == 0xb3) &&
            (sig[4] == 0xb3) &&
            (sig[5] == 0xb0) &&
            (sig[6] == 0xa2) &&
            (sig[7] == 0xb5);
}

/* Return smaller of two numbers. */
int min(int a, int b) {
    if (a<b) 
        return a;
    else
        return b;
}

/* Given minutes since midnight, print the time. */
printTime(winWord minutes) {

    printf("%d:", minutes/60);

    if (minutes % 60 < 10) 
        putchar('0');

    printf("%d", minutes % 60);
}

/* Given days since 1 Jan 1980, print the date. */
printDate(winUint days) {

    char dateString[64];

    /* UNIX uses seconds since 1 January 1970.  There is a difference
     * of 3652 days between the two dates (ten 365-day years plus leap
     * days in 1972 and 1976). Thus, ...  */

    time_t secs = SECS_PER_DAY * (days + 3652);

    strftime(dateString, 63, "%A, %d %B %Y", gmtime(&secs));
    printf("%s", dateString);
}

/* Print the marks that Calendar uses in the whole-month view. */
printSymbols(winWord mark) {

    if (mark != 0) printf(" -- ");

    if (mark & 0x0080) printf("["); /* rectangle */
    if (mark & 0x0100) printf("(");
    if (mark & 0x0200) printf("o");
    if (mark & 0x0400) printf("x");
    if (mark & 0x0800) printf("_");

}

printUsage(char *name) {
    fprintf(stderr, "Usage: %s filename\n", name);
    fprintf(stderr, "    extract information from a Windows Calendar file\n");
}

/*******************************************************************/

main(int argc, char **argv) {

    FILE *fred;
    calHeader myCalHeader;
    int i;

    if (argc != 2) {
        printUsage(argv[0]);
        exit(1);
    }

    if ( (fred = fopen(argv[1], "r")) == NULL) {
        fprintf(stderr, "%s: fopen failed (%d)\n", argv[0]);
        exit(1);
    }

    /* read the header */
    if ( fread(&myCalHeader, sizeof(calHeader), 1, fred) != 1) {
        fprintf(stderr, "%s: fread failed [1]\n", argv[0]);
        exit(1);
    }

    if ( !checkSignature(myCalHeader.signature) ) {
        fprintf(stderr, "%s: file signature check failed\n", argv[0]);
        exit(1);
    }

    for(i=0; i<myCalHeader.dateDescriptors; i++) {

        dateInfo myDateInfo;

        fseek(fred, 0x40 + i*sizeof(dateInfo), SEEK_SET);

        if ( fread(&myDateInfo, sizeof(dateInfo), 1, fred) != 1) {
            fprintf(stderr, "%s: fread failed [2]\n", argv[0]);
            exit(1);
        }

        if (i>0)             /* print a blank line between entries */
            printf("\n");

        printDate(myDateInfo.date);
        printSymbols(myDateInfo.markSymbol);

        if (myDateInfo.fileBlockOffset == 0xffff) 
            printf(" -- no data for this day.\n");
        else {
            dayInfo	myDayInfo;

            printf("\n");

            fseek(fred, 64 * myDateInfo.fileBlockOffset, SEEK_SET);
            if ( fread(&myDayInfo, 10, 1, fred) != 1) {
                fprintf(stderr, "%s: fread failed [3]\n", argv[0]);
                exit(1);
            }

            if ( myDayInfo.date != myDateInfo.date) {
                fprintf(stderr, "%s: day/date mismatch\n", argv[0]);
                exit(1);
            }

            if (myDayInfo.cbNotes > 0) {
                int i;

                printf("    ");

                for (i=0; i<myDayInfo.cbNotes; i++) {
                    char c;
                    if ( fread(&c, 1, 1, fred) != 1) {
                        fprintf(stderr, "%s: fread failed [4]\n", argv[0]);
                        exit(1);
                    }
                    if (c == '\n') 
                        printf("\n    ");
				    else
				        putchar(c);
                }

                printf("\n");

            }

            if (myDayInfo.cbAppointments > 0) {

                /* All appointment records for the day use a total of 
                 * myDayInfo.cbAppointments bytes.  We read records
                 * until we reach this quota, and keep track of bytes
                 * read in cbUsed. */

                winUint cbUsed=0;

                fseek(fred, 64 * myDateInfo.fileBlockOffset + 
                          sizeof(dayInfo) + myDayInfo.cbNotes, SEEK_SET);

                while(cbUsed < myDayInfo.cbAppointments) {

                    apptInfo myApptInfo;

                    if ( fread(&myApptInfo, 4, 1, fred) != 1) {
                        fprintf(stderr, "%s: fread failed [5]\n", argv[0]);
                        exit(1);
                    }

                    cbUsed += myApptInfo.apptSize;

                    if ( myApptInfo.apptSize > 4) {
                        char apptNotes[256];
                        if ( fread(&apptNotes, myApptInfo.apptSize - 4, 1, fred) != 1) {
                            fprintf(stderr, "%s: fread failed [6]\n", argv[0]);
                            exit(1);
                        }
                        apptNotes[255]='\0';
                        printTime(myApptInfo.time);
                        if(myApptInfo.flags > 0) 
                           printf(" alarm");
                        printf(" -- %s\n", apptNotes);
                    }

                } /* while */

            } /* if appointments */

        } /* if data for today */

    } /* for i */

    fclose(fred);
}
